from __future__ import annotations

from dataclasses import dataclass, asdict
from typing import List
from typing import cast

from pydantic import BaseModel

from web.trend.rest.utils.openapi_gen import generate_openapi

try:
    import orjson
    from sanic import Sanic
except ImportError:
    raise ImportError("poetry install --with rest")

from web import settings
from web.kernel.transport import Transport
from web.kernel.types import ConfAble, AppNameAble, ISocket, IService, TransportConf
from web.trend.rest.custom import CustomRouter, CustomErrorHandler, RouterConf


class SanicConf(BaseModel):
    access_log: bool | None = False
    noisy_exceptions: bool | None = False
    debug: bool = False


class SanicExtConf(BaseModel):
    params: SanicConf | None


class RestTransportConfig(TransportConf):
    sanic: SanicExtConf
    router: RouterConf | None


class RestTransport(Transport, AppNameAble, ConfAble[RestTransportConfig]):
    sanic: Sanic
    _router: CustomRouter

    def __init__(self,
                 services: List[IService],
                 routes: dict,
                 ):
        super().__init__(services)
        self.sanic = Sanic("__broken__name",
                           loads=orjson.loads,
                           dumps=orjson.dumps,
                           error_handler=CustomErrorHandler())
        self._router = CustomRouter(routes)
        self._router.ctx.app = self.sanic
        self.sanic.router = self._router

    def generate_openapi_file(self, json_filepath: str | None = None):
        generate_openapi(self.sanic, json_filepath)

    async def run(self, socket: ISocket):
        self.sanic.name = self.app_name
        cast(CustomRouter, self.sanic.router).set_routes(self)

        start_args = self.conf.sanic.params.dict() if self.conf else {}
        if settings.DEBUG:
            start_args["debug"] = True
        server = await self.sanic.create_server(**start_args,
                                                sock=socket,
                                                return_asyncio_server=True)
        await server.startup()
        await server.before_start()
        await server.after_start()
        await server.serve_forever()



import os, os.path
from json import dump, load

import loguru
from sanic import Request, json, Sanic
from sanic_routing import Route
from sanic_ext.config import add_fallback_config
from sanic_ext.extensions.openapi.blueprint import blueprint_factory
from sanic_ext.extensions.openapi.builders import SpecificationBuilder
from sanic_ext.extensions.openapi import extension

from web import settings


def generate_openapi(sanic: Sanic, json_filepath: str | None):
    """
    Generate openapi json
    NOTE: After generation you can't start _sanic app

    """
    settings.OPENAPI = True
    sanic.router.open_api = True

    sanic.router.set_routes(None)
    sanic.finalize()
    sanic.config = add_fallback_config(sanic)
    oas_bp = blueprint_factory(sanic.config)
    for listener in oas_bp._future_listeners:
        if "build_spec" == listener.listener.__name__:
            listener.listener(sanic, None)
    oas_json = SpecificationBuilder().build(sanic).serialize()
    if not json_filepath:
        try:
            if hasattr(sanic.router.conf, "openapi_filepath") and sanic.router.conf.openapi_filepath:
                json_filepath = sanic.router.conf.openapi_filepath
            else:
                raise Exception("openapi_filepath not set in config")
        except RuntimeError:
            raise Exception("openapi_filepath not set in config")
    with open(json_filepath, 'w') as f:
        dump(oas_json, f)


def add_openapi(sanic: Sanic, openapi_filepath: str):
    """
    Overwrite Openapi _sanic extension startup method for reading openapi json from file

    """
    if not openapi_filepath:
        sanic.router.open_api = False
        sanic.config.OAS = False
        return
    if not os.path.exists(openapi_filepath):
        loguru.logger.error('OPENAPI FILE NOT FOUND')
        sanic.router.open_api = False
        sanic.config.OAS = False
        return

    sanic.router.open_api = True
    sanic.config.OAS = True

    async def spec(request: Request):
        with open(openapi_filepath) as f:
            return json(load(f))

    if hasattr(sanic.config, "OAS_URL_PREFIX") and hasattr(sanic.config, "OAS_URI_TO_JSON"):
        for route_path, route in sanic.router.routes_all.items():
            route: Route
            if route.uri == f"{sanic.config.OAS_URL_PREFIX}{sanic.config.OAS_URI_TO_JSON}":
                route.handler = spec

    def startup(self, bootstrap) -> None:
        if self.app.config.OAS:
            self.bp = blueprint_factory(self.app.config)
            for route in self.bp._future_routes:
                if route.uri == self.app.config.OAS_URI_TO_JSON:
                    self.bp._future_routes.remove(route)
                    self.bp.add_route(spec, self.app.config.OAS_URI_TO_JSON)
                    break
            for listener in self.bp._future_listeners:
                if "build_spec" == listener.listener.__name__:
                    self.bp._future_listeners.remove(listener)
            self.app.blueprint(self.bp)
            bootstrap._openapi = SpecificationBuilder()

    extension.OpenAPIExtension.startup = startup



import hashlib
import random
import string
import sys
from typing import Iterable


# from Crypto.Cipher import AES
# from Crypto.Random import get_random_bytes


class BaseCrypto:
    class DataStruct:
        pass

    def __init__(self):
        pass

    @staticmethod
    def encrypt_password(passw: str, salt_len: int = 10) -> Iterable[str]:
        salt = ''.join(random.choices(string.ascii_uppercase + string.digits, k=salt_len))
        crypted_pass = hashlib.sha256((passw + salt).encode(sys.getdefaultencoding())).hexdigest()
        return crypted_pass, salt

    @staticmethod
    def verify_password(raw_pass: str, db_pass: str, salt: str):
        crypted_pass = hashlib.sha256((raw_pass + salt).encode(sys.getdefaultencoding())).hexdigest()
        return db_pass == crypted_pass

    def encrypt(self, plain_text: str) -> DataStruct:
        raise NotImplementedError()

    def decrypt(self, aes_struct: DataStruct) -> str:
        raise NotImplementedError()



import asyncio
import os
from abc import ABCMeta
from multiprocessing.context import BaseContext
from multiprocessing.process import BaseProcess
from signal import SIGINT, SIGTERM
from typing import Dict, Any, Coroutine, Tuple

import loguru

from web import settings
from web.kernel.types import ChanAble, cheat, EnvAble, SharedStateAble, SharedState, ConfAble


class Isolate(ChanAble, EnvAble, SharedStateAble, metaclass=ABCMeta):
    name: str
    init_args: Tuple[Any]
    init_kwargs: Dict[str, Any]
    process: BaseProcess | None
    _coro_stop_condition: str

    def __init__(self, name: str, *args, coro_stop_condition: str = asyncio.FIRST_EXCEPTION, **kwargs):
        self._coro_stop_condition = coro_stop_condition
        self.name = name
        self.init_args = args
        self.init_kwargs = kwargs
        self.process = None

    async def work(self, *args, **kwargs) -> None:
        raise NotImplementedError

    def fork(self, ctx: BaseContext, *args, daemon: bool = True, **kwargs):
        self.process = ctx.Process(target=self.forked, args=args, daemon=daemon, **kwargs)
        self.process.start()
        return self.process

    async def stop(self, ex: Exception = None):
        await self.env.shutdown()
        if settings.DEBUG:
            if ex:
                loguru.logger.exception(ex)
            else:
                loguru.logger.debug(
                    f"{self.name}::Isolate(pid={os.getgid()}) Received terminate signal. Shutting down.")
        exit(0)

    def forked(self, *args, **kwargs) -> None:
        try:

            async def _start():
                loop = asyncio.get_event_loop()
                loop.add_signal_handler(SIGINT, lambda: asyncio.create_task(self.stop()))
                loop.add_signal_handler(SIGTERM, lambda: asyncio.create_task(self.stop()))
                done_tasks, pending_tasks = await asyncio.wait(self.perform(), return_when=self._coro_stop_condition)
                for task in done_tasks:
                    if ex := task.exception():
                        loguru.logger.error(f"[{self.name}::Exception]  {ex}")
                        raise ex

            self.shared_state = SharedState(self.channel, forked=True)
            cheat(self, ChanAble, self.channel)
            cheat(self, SharedStateAble, self.shared_state)
            cheat(self, ConfAble, self.env.conf)
            asyncio.run(self.env.init())
            asyncio.run(_start())
        except Exception as e:
            loguru.logger.exception(e)
            exit(-1)

    def perform(self) -> list[Coroutine[Any, Any, Any] | Coroutine[Any, Any, None]]:
        return [
            self.channel.listen_consume(),
            self.work(*self.init_args, **self.init_kwargs)
        ]

    def __repr__(self):
        return f"{self.name}::Isolate"



import asyncio
from functools import wraps, partial
from inspect import isawaitable, iscoroutine, iscoroutinefunction
from typing import Union, List, Type, Callable, Coroutine, Any


# ------------------------------------------------STR_TO---------------------------------------------------------------


def str_to_list(string: str) -> Union[List[str], None]:
    if not ",":
        return None
    return string.split(",")


def str_to_bool(string: str) -> Union[bool, None]:
    try:
        if string.lower() == "false":
            return False
        elif string.lower() == "true":
            return True
        else:
            return None
    except Exception as e:
        return None


def all_subclasses(cls):
    return set(cls.__subclasses__()).union(
        [s for c in cls.__subclasses__() for s in all_subclasses(c)])


def in_obj_subclasses(obj: object | Type, target_cls: Type):
    try:
        for i in dir(obj):
            if not i.startswith("__"):
                item = getattr(obj, i)
                if item.__class__ in all_subclasses(target_cls):
                    yield item, i
    except Exception as e:
        raise e


def is_in_base_classes(obj: object | Type, target_cls: Type) -> bool:
    try:
        if obj.__class__ == target_cls:
            return True
        return target_cls in obj.__class__.__mro__
    except Exception as e:
        raise e


def in_obj_classes(obj: object | Type, target_cls: Type):
    try:
        for i in dir(obj):
            item = getattr(obj, i)
            if item.__class__ == target_cls:
                yield item, i
    except Exception as e:
        raise e


def in_obj_cls_items(obj: object | Type, target_cls: Type):
    try:
        for i in dir(obj):
            print()
            item = getattr(obj, i)
            cls = getattr(obj, i).__class__
            if item.__class__ == target_cls or cls in all_subclasses(target_cls):
                yield item, i
    except Exception as e:
        raise e


def get_callable_from_fnc(fnc: Callable[..., Coroutine[Any, Any, Any]] | partial):
    caller = fnc
    if iscoroutine(fnc) or isawaitable(fnc):
        @wraps(fnc)
        def __call(*args, **kwargs):
            asyncio.run(fnc(*args, **kwargs))

        caller = __call

    elif iscoroutinefunction(fnc):
        @wraps(fnc)
        def __call(*args, **kwargs):
            asyncio.run(fnc(*args, **kwargs))

        caller = __call

    return caller



from __future__ import annotations

import asyncio
import os
import re
import socket
from abc import ABCMeta
from dataclasses import dataclass
from enum import Enum
from importlib.abc import FileLoader
from importlib.util import spec_from_loader, module_from_spec
from multiprocessing import Queue
from pathlib import Path
from typing import Any, Type, TypeVar, Generic, Callable, Coroutine, get_args, NewType, cast, Tuple, TypeAlias
from typing import Union, Dict, List

import loguru
from pydantic import BaseModel

from web import settings

# --------------------------------------------------GENS TYPES---------------------------------------------------------

JSON = Union[Dict[str, Any], List[Any], int, str, float, bool, Type[None]]
TypeJSON = Union[Dict[str, 'JSON'], List['JSON'], int, str, float, bool, Type[None]]

GenAsyncCall: TypeAlias = Callable[[...], Coroutine[Any, Any, Any]]

GenPerformCall: TypeAlias = list[Coroutine[Any, Any, Any] | Coroutine[Any, Any, None]]

GenConfig: TypeAlias = TypeVar("GenConfig", bound=BaseModel)

DtoType: TypeAlias = TypeVar("DtoType", bound=BaseModel)


# --------------------------------------------------MESSAGING TYPES-----------------------------------------------------


class IMessage:
    message_type: str = "__all__"
    index: int
    inner_index: int
    sender: str
    destination: str = "__all__"
    exec_inner: bool = False
    return_event_type: Type[IMessageReturn] = None

    def __init__(self):
        self.index = 0
        self.inner_index = 0
        self.sender = "None"

    def __str__(self):
        return f"{self.__class__.__name__}({self.message_type},{self.index=},{self.inner_index=}, {self.sender=})"

    def reply(self, *args, **kwargs):
        return self.return_event_type(self, *args, **kwargs)


class IMessageReturn(IMessage):
    def __init__(self, response_to: IMessage, *args, **kwargs):
        super().__init__()
        self.inner_index = response_to.inner_index
        self.destination = response_to.sender


GenEventsType = TypeVar("GenEventsType", bound=List[Type[IMessage] | str] | Type[IMessage] | str)


class IMessageWaiter:
    result: IMessage | None
    wait_for_index: int = 0
    waiter: IChannel
    _timeout: float

    def __init__(self, waiter: IChannel, question: IMessage, timeout: float = 0):
        self.result = None
        self.waiter = waiter
        self.wait_for_index = question.inner_index
        self._timeout = timeout

    async def wait_on(self) -> IMessage | None:
        if settings.DEBUG:
            loguru.logger.debug(f"[{self}] wait for message ")
        _wait_secs = 0
        while self.result is None:
            if self._timeout != 0 and _wait_secs >= self._timeout:
                return None
            await asyncio.sleep(0.05)
            _wait_secs += 0.05
        else:
            return self.result

    def answer(self, message: IMessage):
        if settings.DEBUG:
            loguru.logger.debug(f"{self} got answer {message}")
        self.result = message

    def __eq__(self, other):
        self.wait_for_index = other.wait_for_index

    def __repr__(self):
        return f"[{self.waiter.name}::IMessageWaiter-{self.wait_for_index}]"


GenIMessage = TypeVar("GenIMessage", bound=IMessage, contravariant=True)
EventListener = Callable[[GenIMessage], Coroutine[Any, Any, Any]]


class IChannel:
    read_timeout = 0.01
    name: str
    consume_pipe: Queue
    produce_pipe: Queue
    _listeners: Dict[str, List[EventListener]]
    _response_waiters: Dict[int, IMessageWaiter]
    _inner_index: int

    def __init__(self, idx: int, isolate_name: str):
        self._inner_index = 0
        self.name = f"{isolate_name}::Channel-{idx}"
        # self.debug = debug
        self._listeners = {}
        self._response_waiters = {}
        self.consume_pipe = Queue()
        self.produce_pipe = Queue()

    def __str__(self):
        return f"IChannel({self.name})"

    async def produce(self, msg: IMessage, need_answer: bool = False, answer_timeout: float = 0) -> IMessage | None:
        raise NotImplementedError

    async def sent_to_consume(self, msg: IMessage):
        self.consume_pipe.put(msg)

    async def listen_produce(self, callback: Callable[[GenIMessage], Coroutine]):
        raise NotImplementedError

    async def listen_consume(self):
        raise NotImplementedError

    def add_event_listener(self, event_type: GenEventsType,
                           callback: EventListener, use_nested_classes: bool = False):
        raise NotImplementedError

    def set_waiter(self, waiter: IMessageWaiter):
        self._response_waiters[waiter.wait_for_index] = waiter

    def answer_to_waiters(self, answer: IMessage):
        waiter = self._response_waiters.get(answer.inner_index)
        if waiter:
            waiter.answer(message=answer)
            self._response_waiters.pop(answer.inner_index)


class SettingsChangeEvent(IMessage):
    message_type = "__change_settings_event__"
    name: str
    value: bool

    def __init__(self, name: str, value: bool):
        super().__init__()
        self.name = name
        self.value = value

    def __str__(self):
        return f"{self.__class__.__name__}({self.name})"


# --------------------------------------------------ABLES TYPES------------------------------------------------------

class Able:
    pass


GenAble = TypeVar("GenAble", bound=Able)


class AppNameAble(Able):
    __app_name: str

    @property
    def app_name(self) -> str:
        if self.__app_name:
            return self.__app_name
        else:
            raise RuntimeError(f"__app_name not set in {self.__class__.__name__}")

    @app_name.setter
    def app_name(self, name: str):
        self.__app_name = name


class SharedStateAble(Able):
    __shared_state: SharedState | None = None

    @property
    def shared_state(self) -> SharedState:
        if self.__shared_state:
            return self.__shared_state
        else:
            raise RuntimeError(f"__shared_state not set in {self.__class__.__name__}")

    @shared_state.setter
    def shared_state(self, state: SharedState):
        self.__shared_state = state


class ChanAble(Able):
    name: str
    __chan: IChannel | None = None

    @property
    def channel(self) -> IChannel:
        if self.__chan:
            return self.__chan
        else:
            raise RuntimeError(f"__channel not set in {self.__class__.__name__}")

    @channel.setter
    def channel(self, chan: IChannel):
        self.__chan = chan


class ConfAble(Generic[GenConfig], Able):
    __conf: GenConfig | None = None

    @property
    def conf(self) -> GenConfig:
        if self.__conf:
            return self.__conf
        else:
            raise RuntimeError("__conf not set")

    @conf.setter
    def conf(self, conf: GenConfig):
        self.__conf = conf


class EnvAble(Able):
    __env: Environment | None

    @property
    def env(self) -> Environment:
        if self.__env:
            return self.__env
        else:
            raise RuntimeError(f"__env not set in {self.__class__.__name__}")

    @env.setter
    def env(self, e: Environment):
        self.__env = e


def cheat(set_to: Any, typed: Type, value: Any, nested: bool = True, self: bool = True, nested_count: int = 1):
    def _set_conf_able(obj, conf) -> bool:
        for base in obj.__orig_bases__:
            if "ConfAble" in base.__name__:
                conf_class = get_args(base)[0]
                for attr in vars(conf).values():
                    if isinstance(attr, conf_class):
                        obj.conf = attr
                        loguru.logger.success(
                            f"[CONF_ABLE::{obj.__class__.__name__}] Set {attr.__class__.__name__}")
                        return True
                    elif issubclass(attr.__class__, BaseModel):
                        if _set_conf_able(obj, attr):
                            return True
            if base == Generic[GenConfig]:
                obj.conf = value
                return True
        return False

    def set_up(target: GenAble):
        if typed == EnvAble:
            target.env = value
            return True
        elif typed == AppNameAble:
            target.app_name = value
            return True
        elif typed == ChanAble:
            target.channel = value
            return True
        elif typed == SharedStateAble:
            target.shared_state = value
            return True
        if typed == ConfAble:
            return _set_conf_able(target, value)
        return False

    def proc(_item):
        if typed in _item.__class__.__mro__:
            if settings.DEBUG:
                loguru.logger.debug(
                    f"[CYCLE::{nested_count}][OBJECT::{set_to.__class__.__name__}] "
                    f"- Set({typed.__name__},{name}:{item.__class__.__name__})")
            is_set = set_up(_item)
            if settings.DEBUG:
                if is_set:
                    loguru.logger.success(
                        f"[CYCLE::{nested_count}][OBJECT::{_item.__class__.__name__}] Set {typed.__name__}")
                else:
                    loguru.logger.error(
                        f"[CYCLE::{nested_count}][OBJECT::{_item.__class__.__name__}] NOT SET {typed.__name__}")
            if nested:
                cheat(_item, typed, value,
                      nested_count=nested_count + 1)

    if self and typed in set_to.__class__.__mro__:
        set_up(set_to)

    if hasattr(set_to, "__dict__"):
        for name, item in vars(set_to).items():
            if isinstance(item, list):
                for i in item:
                    proc(i)
            else:
                proc(item)
    else:
        loguru.logger.error(f"SKIP {set_to.__class__.__name__} have not __dict__")


# --------------------------------------------------SIGNALS TYPES-------------------------------------------------------

class SignalType(Enum):
    BEFORE_APP_RUN = "BEFORE_APP_RUN"
    BEFORE_TRANSPORT_WORK = "BEFORE_TRANSPORT_WORK"
    AFTER_APP_STOP = "AFTER_APP_STOP"


class SigAble:
    signals: Dict[SignalType, List[GenAsyncCall]] | None = None

    def on_signal(self, typed: SignalType, func: GenAsyncCall):
        if not self.signals:
            self.signals = {}
        if not self.signals.get(typed):
            self.signals[typed] = []
        self.signals[typed].append(func)


async def call_signals(obj: SigAble, typed: SignalType, *args, nested: bool = False, **kwargs):
    if hasattr(obj, "signals") and obj.signals and obj.signals.get(typed):
        for signal in obj.signals[typed]:
            await signal(*args, **kwargs)
        if nested:
            for item in vars(obj).values():
                if SigAble in type(item).__mro__:
                    await call_signals(item, typed, *args, nested=True, **kwargs)


# --------------------------------------------------PROC TYPES---------------------------------------------------------


class SetStateEvent(IMessage):
    message_type = "__set_state__"
    key: Any
    value: Any
    destination = "__master__"
    exec_inner = False

    def __init__(self, key: Any, value: Any):
        super().__init__()
        self.key = key
        self.value = value


class GetStateReturnEvent(IMessageReturn):
    message_type = "__get_state_return__"
    exec_inner = False
    value: Any

    def __init__(self, ret_to: GetStateEvent, value: Any):
        super().__init__(ret_to)
        self.value = value


class GetStateEvent(IMessage):
    message_type = "__get_state__"
    key: Any
    destination = "__master__"
    exec_inner = False
    return_event_type = GetStateReturnEvent

    def __init__(self, key: Any):
        super().__init__()
        self.key = key


class RemoveStateEvent(IMessage):
    message_type = "__remove_state__"
    key: Any
    destination = "__master__"
    exec_inner = False

    def __init__(self, key: Any):
        super().__init__()
        self.key = key


class SharedState(ChanAble):
    _state: Dict[str, Any]
    locally: bool
    sync: asyncio.Lock
    forked: bool
    set_event_type: Type[IMessage] = SetStateEvent
    get_event_type: Type[IMessage] = GetStateEvent
    remove_event_type: Type[IMessage] = RemoveStateEvent

    def __init__(self, channel: IChannel,
                 forked=False,
                 locally: bool = False):
        self._state = {}
        self.channel = channel
        self.locally = locally
        self.forked = forked
        self.sync = asyncio.Lock()
        if not forked:
            if settings.DEBUG:
                loguru.logger.debug(f"{self} Set events listeners")
            self.channel.add_event_listener(self.set_event_type, self._set)
            self.channel.add_event_listener(self.get_event_type, self._get)
            self.channel.add_event_listener(self.remove_event_type, self._remove)

    async def _remove(self, event: RemoveStateEvent):
        async with self.sync:
            if settings.DEBUG:
                loguru.logger.debug(f"{self} Remove state Key: {event.key}")
            self._state.pop(event.key)

    async def _set(self, event: SetStateEvent):
        async with self.sync:
            if settings.DEBUG:
                loguru.logger.debug(f"{self} Set state Key: {event.key} Value: {event}")
            self._state[event.key] = event.value

    async def _get(self, event: GetStateEvent):
        async with self.sync:
            if settings.DEBUG:
                loguru.logger.debug(f"{self} Get state Key: {event.key}")
            await self.channel.produce(event.reply(self._state.get(event.key)))

    async def set(self, key: Any, value: Any) -> None:
        if self.locally:
            self._state[key] = value
        await self.channel.produce(self.set_event_type(key, value))

    async def get(self, key: Any):
        if self.locally and key in self._state:
            return self._state[key]
        response = cast(GetStateReturnEvent, await self.channel.produce(self.get_event_type(key), need_answer=True))
        return response.value

    async def remove(self, key: Any):
        if self.locally and key in self._state:
            self._state.pop(key)
        await self.channel.produce(self.remove_event_type(key))

    def __repr__(self):
        return f"[{self.__class__.__name__}::{self.channel.name}]"

    @property
    def local_store(self) -> Dict[str, Any]:
        if self.locally:
            return self._state
        else:
            raise RuntimeError(f"{self} Local store unavailable, locally is false")


# --------------------------------------------------PLUGINS TYPES ------------------------------------------------------


class Plugin(AppNameAble, EnvAble, SigAble, SharedStateAble, metaclass=ABCMeta):
    name: str
    enabled: bool
    required_attrs: List[str] = []

    async def exec(self, *args, **kwargs):
        raise NotImplementedError

    def __repr__(self):
        return f"{self.__class__.__name__}"


class PluginContainer:
    _REGEX = re.compile(r"PLUGIN_NAME\s?=\s?[\"|'](.*)[\"|']")
    file_path: Path
    code: str
    plugin: Plugin | None

    class SourceCodeLoader(FileLoader):
        def __init__(self, fullname: str, source):
            super().__init__(fullname, source)
            self.path = source

        def get_source(self, fullname: str) -> str | bytes:
            return self.path

    def __init__(self, file_path: Path):
        self.file_path = file_path
        self.plugin = None

    def import_plugin(self) -> Tuple[str, str | Exception]:
        if not self.file_path.exists():
            return "error", FileExistsError(f"{self.file_path} does not exist")
        with open(self.file_path) as file:
            self.code = file.read()
            match = self._REGEX.findall(self.code)
            if match:
                try:
                    plugin_name = match[0]
                    spec = spec_from_loader(self.file_path.name,
                                            loader=self.SourceCodeLoader(plugin_name, self.code))
                    module = module_from_spec(spec)
                    spec.loader.exec_module(module)
                    self.plugin = getattr(module, plugin_name)()
                    if Plugin not in self.plugin.__class__.__mro__:
                        return "warning", f"{plugin_name} is not child on Plugin"
                    self.plugin.name = plugin_name
                    for at in self.plugin.required_attrs:
                        if not hasattr(self.plugin, at) and not getattr(self.plugin, at):
                            return "warning", f"{at} not found in {self.plugin.name}"
                    return "success", f"Import plugin"
                except Exception as e:
                    return "exception", e
            return "warning", f"No plugin name found in {self.file_path} {match}"

    def __repr__(self):
        return f"{self.__class__.__name__}::{self.plugin}"

    def __getstate__(self):
        self.plugin = None
        return self.__dict__


class SetPluginEvent(SetStateEvent):
    message_type = "__set_plugin_state__"
    key: str
    destination: str = "__all__"
    value: PluginContainer
    exec_inner = True

    def __init__(self, key: str, value: PluginContainer):
        super().__init__(key, value)


class RemovePluginEvent(RemoveStateEvent):
    message_type = "__remove_plugin_state__"
    destination: str = "__all__"
    exec_inner = True


class PluginsStore(SharedState, EnvAble):
    _state: Dict[str, PluginContainer]
    dir_path: Path
    set_event_type: Type[IMessage] = SetPluginEvent
    remove_event_type: Type[IMessage] = RemovePluginEvent

    def __init__(self, dir_path: Path, channel: IChannel, *args, **kwargs):
        self.dir_path = dir_path
        super().__init__(channel, *args, locally=True, forked=True, **kwargs)
        self.channel.add_event_listener(self.set_event_type, self._set)
        self.channel.add_event_listener(self.remove_event_type, self._remove)

    def import_plugins(self):
        for filename in os.listdir(self.dir_path):
            plugin_path = self.dir_path.joinpath(str(filename))
            if plugin_path.is_dir():
                continue
            container = PluginContainer(plugin_path)
            msg_type, message = container.import_plugin()
            if settings.DEBUG:
                getattr(loguru.logger, msg_type)(f"[{self}::{container}] {message}")
            if msg_type == "success":
                self._state.update({container.plugin.name: container})

    async def _set(self, event: SetPluginEvent):
        await super()._set(event)
        status, message = event.value.import_plugin()
        if settings.DEBUG:
            getattr(loguru.logger, status)(f"[{self}::{event.value}] {message}")

    async def get(self, key: str) -> PluginContainer:
        return cast(PluginContainer, await super().get(key))

    async def set(self, key: Any, value: Any) -> None:
        await super().set(key, value)

    def __repr__(self):
        return f"{self.__class__.__name__}({os.getpid()})"

    @property
    def local_store(self) -> Dict[str, PluginContainer]:
        return super().local_store


# --------------------------------------------------SPECIAL TYPES------------------------------------------------------

class SocketConf(BaseModel):
    host: str | None
    port: int | None


class ISocket(socket.socket):
    host: int
    port: int
    backlog: int

    def __init__(self, host, port, backlog: int = 100):
        """Create TCP server socket.
        :param host: IPv4
        :param port: TCP port number
        :param backlog: Maximum number of connections to queue
        """
        super().__init__()
        self.host = host
        self.port = port
        self.backlog = backlog
        self.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        self.bind((self.host, self.port))
        self.listen(self.backlog)
        self.set_inheritable(True)

    def __repr__(self):
        return f"{self.host}:{self.port}"


GenISocket = TypeVar("GenISocket", bound=ISocket)

# --------------------------------------------------MAIN TYPES---------------------------------------------------------

GenCommunicator = TypeVar("GenCommunicator")


class TransportConf(BaseModel):
    socket: SocketConf


class ITransport(ConfAble[TransportConf], ChanAble, EnvAble, SharedStateAble, SigAble, metaclass=ABCMeta):
    services: List[GenService]
    rt_writers: Dict[Type[RtWriter], List[RtWriter]]
    plugins_dir: Path | None
    _plugins_store: PluginsStore | None

    def __init__(self, services: List[GenService]):
        EnvAble.__init__(self)
        self.services = services
        for service in self.services:
            service.transport = self
        self.rt_writers = {}
        self._plugins_store = None
        self.plugins_dir = None

    def get_service(self, typed: Type[GenService]) -> GenService | None:
        for service in self.services:
            if isinstance(service, typed):
                return service

    def with_plugins(self, plugins_dir: Path):
        """
        enable plugins in the transport
        """
        if not plugins_dir.exists():
            raise FileNotFoundError(f"{plugins_dir} does not exist")
        self.plugins_dir = plugins_dir
        return self

    async def set_settings(self, name: str, value: Any):
        setattr(settings, name, value)
        await self.channel.produce(SettingsChangeEvent(name, value))

    def init_plugins(self):
        if self.plugins_dir:
            self._plugins_store = PluginsStore(self.plugins_dir, self.channel)
            self._plugins_store.import_plugins()

    def perform(self, sock: ISocket) -> GenPerformCall:
        self.init_plugins()
        return [self.run(sock)]

    async def back_task(self,
                        target: BackTaskTarget,
                        args=None,
                        kwargs=None,
                        trigger: BackTaskTrigger = None,
                        exec_type: str = "default",
                        on_error: BackTaskCallback = None,
                        on_complete: BackTaskCallback = None,
                        need_result: bool = False
                        ) -> TaskReturnEvent | None:
        if args is None:
            args = []
        if kwargs is None:
            kwargs = {}
        raise NotImplementedError

    async def new_rt(self, writer: RtWriter,
                     events_type: GenEventsType,
                     resolver: RtResolver,
                     use_nested_events: bool = False):
        raise NotImplementedError

    async def run(self, sock: ISocket):
        raise NotImplementedError

    @property
    def plugins(self):
        if not self._plugins_store:
            return []
        return [container.plugin for container in self._plugins_store.local_store.values()]


class Resource(AppNameAble, ConfAble, ChanAble, metaclass=ABCMeta):
    __communicator: GenCommunicator | None

    def __init__(self, communicator: GenCommunicator | None = None):
        self.__communicator = communicator

    async def init(self, *args, **kwargs) -> None:
        raise NotImplementedError

    async def shutdown(self):
        raise NotImplementedError

    @property
    def communicator(self):
        if not self.__communicator:
            raise RuntimeError(f"__communicator not set in {self.__class__.__name__}")
        return self.__communicator

    @communicator.setter
    def communicator(self, comm: GenCommunicator):
        self.__communicator = comm


class Environment(ConfAble, AppNameAble, ChanAble, SharedStateAble):
    _resources: List[Resource]

    def __init__(self):
        self._resources = []
        for res in self.__class__.__dict__.values():
            if Resource in res.__class__.__mro__:
                self._resources.append(res)

    async def init(self):
        for res in self._resources:
            cheat(res, AppNameAble, self.app_name, nested=False)
            cheat(res, ConfAble, self.conf, nested=False)
            cheat(res, ChanAble, self.channel, nested=False)
            await res.init()

    async def shutdown(self):
        for resource in self._resources:
            await resource.shutdown()


class IService:
    transport: ITransport


GenService = TypeVar("GenService", bound=IService)


# --------------------------------------------------RT TYPES -----------------------------------------------------------


@dataclass
class RtMessage:
    def prepare(self, idx: int = 0, *args, **kwargs) -> str | bytes:
        raise NotImplementedError

    @classmethod
    def ping_message(cls):
        raise NotImplementedError


GenRtMessage = TypeVar("GenRtMessage", bound=RtMessage)


class RtWriter(Generic[GenRtMessage]):
    message_class: GenRtMessage
    last_index: int = 0
    obj: Any
    on_disconnect: List[GenAsyncCall] = None

    def __init__(self, obj: Any):
        self.obj = obj
        self.on_disconnect = []

    def on_close(self):
        for task in self.on_disconnect:
            asyncio.create_task(task())

    async def write(self, message: GenIMessage) -> None:
        raise NotImplementedError

    async def freeze(self):
        raise NotImplementedError

    def with_lost_callback(self, on_disconnect: GenAsyncCall) -> RtWriter:
        self.on_disconnect.append(on_disconnect)
        return self

    async def run(self):
        await self.freeze()

    async def close(self, *args, **kwargs):
        raise NotImplementedError


RtResolver = NewType("RtResolver", Callable[[RtWriter, GenIMessage], Coroutine[Any, Any, RtMessage | None]])
RtReadCallback = NewType("RtReadCallback", Callable[[RtWriter, Any], Coroutine[Any, Any, RtMessage | None]])

# --------------------------------------------------BACK TYPES ---------------------------------------------------------


BackTaskTarget = NewType("BackTaskTarget", Callable[[Environment, ...], Coroutine[Any, Any, Any]])
BackTaskTrigger = TypeVar("BackTaskTrigger")
BackTaskCallback = TypeVar("BackTaskCallback")

ISOLATE_EXEC = "isolate_exec"


class TaskReturnEvent(IMessageReturn):
    message_type = "__task_return__"
    result: Any
    ex: Exception

    def __init__(self, ret_to: TaskEvent, result: Any, ex: Exception = None):
        super().__init__(ret_to)
        self.result = result
        self.ex = ex


class TaskEvent(IMessage):
    message_type = "__task__"
    destination = "__master__"
    target: BackTaskTarget
    trigger: BackTaskTrigger
    exec_type: str
    args: List[Any]
    kwargs: Dict[str, Any]
    on_error: BackTaskCallback
    on_complete: BackTaskCallback
    return_event_type = TaskReturnEvent

    def __init__(self, target: BackTaskTarget,
                 trigger: BackTaskTrigger = None,
                 args: List[Any] = None,
                 exec_type: str = "default",
                 kwargs: Any | None = None,
                 on_error: BackTaskCallback = None,
                 on_complete: BackTaskCallback = None):
        super().__init__()
        self.target = target
        self.args = args
        self.kwargs = kwargs
        self.trigger = trigger
        self.exec_type = exec_type
        self.on_error = on_error
        self.on_complete = on_complete


class IScheduler:
    _tasks_info: Dict[str, BackTaskTarget]
    state: Dict[str, Any]

    async def add_task(self, message: GenIMessage):
        raise NotImplementedError

    def stop(self, *args, **kwargs):
        raise NotImplementedError

    def perform(self) -> List[Coroutine[Any, Any, Any] | Coroutine[Any, Any, None]]:
        raise NotImplementedError



from typing import TypeVar, Generic

# PLUGIN_NAME = "SHITPOSTLIGIN"
#
# class SHITPOSTLIGIN:
#     pass
#
# class T:
#     def w(self):
#         pass
#
#
# GenT = TypeVar("GenT", bound=T)
#
#
# class ChanAble(Generic[GenT]):
#     _chan: GenT | None = None
#
#     @property
#     def channel(self) -> GenT:
#         if self._chan:
#             return self._chan
#         else:
#             raise Exception("_channel not set")
#
#     @channel.setter
#     def channel(self, chan: GenT):
#         self._chan = chan
#
#
# class MyT(T):
#
#     def c(self):
#         pass
#
#
# GenM = TypeVar("GenT")
#
#
# class Test(Generic[GenM]):
#     def __init__(self):
#         from typing import get_args
#         print(self.__orig_bases__)
#         print(get_args(self.__orig_bases__))
#
#
# if __name__ == '__main__':
#     # Test[MyT]().channel.c()
#     t = Test[MyT]()

# import asyncio
# from typing import Dict, List
#
# from news.kernel.signal import SignalType
# from news.kernel.types import GenAsyncCall
#
#
# class SignalAble:
#     __signals: Dict[SignalType, List[GenAsyncCall]] | None = None
#
#     def __init__(self):
#         self.on_signal(typed=SignalType.BEFORE_APP_RUN, func=self.test)
#
#     def on_signal(self, typed: SignalType, func: GenAsyncCall):
#         if not self.__signals:
#             self.__signals = {}
#         if not self.__signals.get(typed):
#             self.__signals[typed] = []
#         self.__signals[typed].append(func)
#
#     async def call_signals(self, typed: SignalType, *args, **kwargs):
#         for signal in self.__signals[typed]:
#             await signal(*args, **kwargs)
#
#     async def test(self):
#         print("pidor")
#
#
# async def main():
#     s = SignalAble()
#     await s.call_signals(SignalType.BEFORE_APP_RUN)
#
#
# if __name__ == '__main__':
#     asyncio.run(main())



class Service:
    def set(self, value):
        pass

class P():
    service_method = Service.set

if __name__ == '__main__':
    print(P().service_method.__func__.__qualname__)
    print(dir(P().service_method.__func__))
    print(P().service_method.__name__)



from __future__ import annotations
from __future__ import annotations

import asyncio
import datetime
import uuid
from dataclasses import dataclass, field
from datetime import timezone
from typing import Callable, Coroutine, Any, List
from typing import Dict

import loguru

from web import settings
from web.kernel.messaging.dispatcher import IDispatcher
from web.kernel.proc.isolate import Isolate
from web.kernel.proc.manager import ProcManager
from web.kernel.types import IScheduler, EnvAble, ISOLATE_EXEC, TaskEvent

if settings.SCHEDULER_ENABLE:
    from apscheduler.events import EVENT_JOB_SUBMITTED, JobSubmissionEvent, EVENT_JOB_ERROR, EVENT_JOB_EXECUTED
    from apscheduler.events import JobExecutionEvent
    from apscheduler.executors.asyncio import AsyncIOExecutor
    from apscheduler.executors.base import BaseExecutor
    from apscheduler.schedulers.base import BaseScheduler
    from apscheduler.jobstores.memory import MemoryJobStore
    from apscheduler.schedulers.asyncio import AsyncIOScheduler
    from apscheduler.triggers.interval import IntervalTrigger


@dataclass
class Task:
    id: str
    status: str
    trigger_event: TaskEvent
    call_time: datetime.datetime | None = field(default=None)
    done_time: datetime.datetime | None = field(default=None)
    execution_time: datetime.timedelta | None = field(default=None)

    error: BaseException | None = field(default=None)
    create_time: datetime.datetime | None = field(default_factory=datetime.datetime.now)

    call_counter: int = field(default=0)

    on_error_callback: TaskErrorCallback = field(default=None)
    on_complete_callback: TaskCompleteCallback = field(default=None)

    def __eq__(self, other: Task):
        if self.trigger_event.target.__name__ == other.trigger_event.target.__name__:
            return True
        return False

    def __repr__(self):
        return self.trigger_event.target.__name__


TaskErrorCallback = Callable[[JobExecutionEvent, Task], Coroutine[Any, Any, None]]
TaskCompleteCallback = Callable[[JobExecutionEvent, Task], Coroutine[Any, Any, None]]


class BackGroundIsolate(Isolate):
    def __init__(self, name: str, *args, coro_stop_condition: str = asyncio.FIRST_EXCEPTION, **kwargs):
        super().__init__(name, *args, coro_stop_condition=coro_stop_condition, **kwargs)

    def forked(self, *args, **kwargs) -> None:
        try:
            asyncio.run(self.env.shutdown())
        except Exception:
            pass
        super().forked(*args, **kwargs)

    async def work(self, *args, **kwargs) -> None:
        await args[0](self.env, *args[1:], **kwargs)


class Scheduler(IScheduler, EnvAble):
    _tasks_info: Dict[str, Task]
    _scheduler: BaseScheduler
    _executors: Dict[str, BaseExecutor]
    manager: ProcManager | None
    dispatcher: IDispatcher | None

    def __init__(self, scheduler: BaseScheduler = None):
        self.manager = None
        self.dispatcher = None
        self._tasks_info = {}
        self._scheduler = scheduler if scheduler else self._default_scheduler()

    def _default_scheduler(self) -> AsyncIOScheduler:
        scheduler = AsyncIOScheduler(jobstores={"default": MemoryJobStore()},
                                     executors={
                                         'default': AsyncIOExecutor(),
                                     },
                                     job_defaults={
                                         "misfire_grace_time": 1,
                                         'coalesce': False,
                                         'max_instances': 10
                                     },
                                     timezone=timezone.utc)

        def _on_job_submitted(event: JobSubmissionEvent):
            nonlocal self
            task = self._tasks_info.get(event.job_id)
            task.call_time = datetime.datetime.now().astimezone()
            task.status = "called"

        def _on_job_exec(event: JobExecutionEvent):  # todo remove done jobs? YAROHA: NO, NEED TO STATISTICS
            task = self._tasks_info.get(event.job_id)
            task.done_time = datetime.datetime.now().astimezone()
            task.execution_time = task.done_time - task.call_time
            task.status = "done"
            if event.exception:
                task.error = event.exception
                task.status = 'error'
                if settings.DEBUG:
                    loguru.logger.error(
                        f"[Scheduler::Task::{task.id}] Job Errored in {task.execution_time} with {event.exception}")
                if task.on_error_callback:
                    asyncio.create_task(task.on_error_callback(event, task))
                # self.manager.remove_isolate(task.name)
            else:
                if settings.DEBUG:
                    loguru.logger.success(f"[Scheduler::Task::{task.id}] Job {task} executed in {task.execution_time}")
                if task.on_complete_callback:
                    asyncio.create_task(task.on_complete_callback(event, task))
            asyncio.create_task(
                self.env.channel.produce(task.trigger_event.reply(event.retval, event.exception)))

        scheduler.add_listener(_on_job_submitted, EVENT_JOB_SUBMITTED)
        scheduler.add_listener(_on_job_exec, EVENT_JOB_EXECUTED | EVENT_JOB_ERROR)
        return scheduler

    async def add_task(self, message: TaskEvent):
        arc_task = Task(id=str(uuid.uuid4()),
                        status="new",
                        trigger_event=message)

        if message.exec_type == ISOLATE_EXEC:
            if not self.manager:
                raise ValueError("manager not set, process call unavailable")

            async def run_in_process(*args, **kwargs):
                isolate = BackGroundIsolate(arc_task.id,
                                            *args,
                                            coro_stop_condition=asyncio.FIRST_COMPLETED,
                                            **kwargs)
                isolate.channel = self.env.channel
                self.dispatcher.set_channel(isolate)
                self.manager.add_isolate(isolate, run=True)
                while isolate.process.is_alive():
                    await asyncio.sleep(0.1)
                isolate.process.terminate()
                self.manager.remove_isolate(isolate.name)
                return

            message.args = (message.target, *message.args)
            run_in_process.__name__ = message.target.__name__
            message.target = run_in_process
        else:
            message.args = (self.env, *message.args)
        if isinstance(message.trigger, IntervalTrigger):
            for task in self._tasks_info.values():
                if task == arc_task:
                    task.trigger_event = message
                    job = self._scheduler.get_job(task.id)
                    job.func = message.target
                    job.trigger = message.trigger
                    job.args = message.args
                    job.kwargs = message.kwargs
                    return

        job = self._scheduler.add_job(message.target,
                                      id=arc_task.id,
                                      name=arc_task.id,
                                      trigger=message.trigger,
                                      args=message.args,
                                      kwargs=message.kwargs)
        if settings.DEBUG:
            loguru.logger.debug(f"[Scheduler::Task::{arc_task.id}] Job {arc_task} queued")
        self._tasks_info.update({job.id: arc_task})

    def stop(self, *args, **kwargs):
        self._scheduler.shutdown()

    async def run(self):
        self._scheduler.start()

    def perform(self) -> List[Coroutine[Any, Any, Any] | Coroutine[Any, Any, None]]:
        self._scheduler.start()

        async def call():
            self._scheduler.start()

        return [call()]



import importlib
import os
from pathlib import Path
from typing import List

from loguru import logger
from pydantic import BaseModel
from tortoise import Tortoise, BaseDBAsyncClient, connections

from web.kernel.types import ConfAble, Resource


class DbConfig(BaseModel):
    host: str
    port: str
    database: str
    user: str
    password: str
    db_schema: str
    with_migrations: bool
    migrations_path: Path


class DatabaseResource(Resource, ConfAble[DbConfig]):
    communicator: BaseDBAsyncClient
    _modules: List[str]
    _engine: str

    def __init__(self, modules: List[str], engine: str = 'tortoise.backends.asyncpg', *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._engine = engine
        self._modules = modules

    def _get_connection_setting(self) -> dict:
        if self.conf.with_migrations and "aerich.models" not in self._modules:
            self._modules.append("aerich.models")
        to_discover = [importlib.import_module(i) for i in self._modules]
        return {
            'connections': {
                # Dict format for connection
                f'{self.app_name}_default': {
                    'engine': self._engine,
                    'credentials': {
                        'host': self.conf.host,
                        'port': self.conf.port,
                        'user': self.conf.user,
                        'password': self.conf.password,
                        'database': self.conf.database,
                        'schema': self.conf.db_schema,
                        'minsize': 1,
                        'maxsize': 5,
                    }
                }
            },
            'apps': {
                f'{self.app_name}': {
                    'models': to_discover,
                    'default_connection': f'{self.app_name}_default',
                }
            },
            'use_tz': False,
            'timezone': 'UTC'
        }

    async def fill_db_data(self):
        pass

    async def init(self, **kwargs):
        await Tortoise.init(config=self._get_connection_setting())
        self.communicator = Tortoise.get_connection(f'{self.app_name}_default')

    async def shutdown(self) -> None:
        await connections.close_all()

    async def _migrations(self, schema_exists: bool, command):
        path_exists = os.path.exists(os.path.join(os.getcwd(), self.conf.migrations_path)) and os.listdir(
            os.path.join(os.getcwd(), self.conf.migrations_path))

        if not path_exists and not schema_exists:
            await self._create_schema(False)
            await command.init()
            await command.init_db(safe=True)
        elif not schema_exists:
            await self._create_schema(False)
            await command.init()
            await command.upgrade()
        elif not path_exists:
            await command.init()
            await command.init_db(safe=True)  # TODO check - need to drop aerich table in db?
        await command.init()
        logger.info(f"Apply migrations from {self.conf.migrations_path}")
        await command.migrate()
        await command.upgrade()

    async def configure_db(self):
        scheme_name = self.conf.db_schema
        row_count, rows = await self.communicator.execute_query(
            "SELECT schema_name FROM information_schema.schemata WHERE schema_name = $1", [scheme_name])
        schema_exists = True if row_count else False
        if self.conf.with_migrations:
            if not self.conf.migrations_path:
                raise ValueError("Migration Path not set can't migrate")
            try:
                from aerich import Command
            except ImportError:
                raise ImportError("To use migration need to install aerich, (web-foundation[aerich])")
            command = Command(tortoise_config=self._get_connection_setting(), app=self.app_name,
                              location=self.conf.migrations_path)
            await self._migrations(schema_exists, command)
        if not schema_exists:
            await self._create_schema()
            await self.fill_db_data()

    async def _create_schema(self, generate_schemas: bool = True):
        await self.communicator.execute_script(f"CREATE SCHEMA IF NOT EXISTS {self.conf.db_schema};")
        if generate_schemas:
            await Tortoise.generate_schemas()



from __future__ import annotations

import asyncio
import json
import os
from pathlib import Path
from signal import Signals, signal, SIGINT, SIGTERM
from typing import List, Generic, Any, Coroutine, Type, cast

import loguru

from web import settings
from web.kernel.messaging.dispatcher import IDispatcher
from web.kernel.proc.manager import ProcManager
from web.kernel.transport import Transport, TransportIsolate
from web.kernel.types import ConfAble, GenConfig, AppNameAble, Environment, EnvAble, cheat, ISocket, SigAble, \
    IScheduler, ChanAble, SocketConf, TaskEvent, SharedState, SignalType, call_signals


class WebApp(Generic[GenConfig], ConfAble[GenConfig], SigAble, ChanAble):
    """
    Main application class.
    """
    _transports: List[Transport]
    name: str
    _environment: Environment
    manager: ProcManager
    scheduler: IScheduler | None
    dispatcher: IDispatcher
    shared_state: SharedState | None

    def __init__(self, name: str,
                 config: GenConfig,
                 transports: List[Transport],
                 environment: Environment):
        self.scheduler = None
        self.shared_state = None
        self._environment = environment
        self._transports = transports
        self.manager = ProcManager()
        self.dispatcher = IDispatcher()
        cheat(self.manager, EnvAble, self._environment, nested=False)
        self.channel = self.dispatcher.set_channel(self, ext_name="MasterChannel", master=True)
        self.name = name
        self.conf = config

    def with_scheduler(self, scheduler: IScheduler = None) -> WebApp:
        """
        add background task scheduler to app.
        """
        if settings.SCHEDULER_ENABLE:
            if not self.shared_state:
                self.shared_state = SharedState(self.channel)
            from web.kernel.proc.scheduler import Scheduler
            self.scheduler = Scheduler() if not scheduler else scheduler
            if not self.scheduler.manager:
                self.scheduler.manager = self.manager
            if not self.scheduler.dispatcher:
                self.scheduler.dispatcher = self.dispatcher
            return self
        else:
            raise RuntimeError("apscheduler is not installed on your system. scheduler not enable")

    def with_shared_state(self) -> WebApp:
        """
        add shared state to app.
        :return:
        """
        self.shared_state = SharedState(self.channel)
        return self

    def _create_transport_isolates(self, workers: int | None = None) -> List[TransportIsolate]:
        isolates = []
        typed_transports: dict[Type[Transport], List[Transport]] = {}
        type_sockets: dict[Type[Transport], ISocket] = {}
        for transport in self._transports:
            if type(transport) not in typed_transports:
                typed_transports.update({type(transport): [transport]})
            else:
                typed_transports[type(transport)].append(transport)
        for typed, transport in typed_transports.items():
            conf = cast(SocketConf, transport[0].conf.socket)
            type_sockets.update({typed: ISocket(conf.host, conf.port)})

        if not workers:
            workers = 0
            for typed, transports in typed_transports.items():
                for num, transport in enumerate(transports):
                    isolate = TransportIsolate(f"{self.name}::{transport.__class__.__name__}-{num}",
                                               transport, type_sockets[typed])
                    self.dispatcher.set_channel(isolate)
                    isolates.append(isolate)
                    workers += 1
        else:
            if workers < len(typed_transports):
                raise Exception(f"Too many transports ({len(typed_transports)}) for this cpu ({workers})")
            workers_count: dict[Type[Transport], int] = {typed: 0 for typed in typed_transports.keys()}
            for typed in typed_transports.keys():
                workers_count.update({typed: 0})
            couter = 0
            from itertools import cycle
            type_cycle = cycle(typed_transports.keys())
            isolates_count: dict[Type[Transport], int] = {typed: 0 for typed in typed_transports.keys()}
            while couter != workers:
                typed = next(type_cycle)
                transports = typed_transports[typed]
                if len(transports) <= workers_count[typed]:
                    workers_count[typed] += 1
                else:
                    workers_count[typed] = 0
                transport = transports[workers_count[typed]]
                isolates_count[typed] += 1
                isolate = TransportIsolate(f"{self.name}::{transport.__class__.__name__}-{isolates_count[typed]}",
                                           transport,
                                           type_sockets[typed])
                self.dispatcher.set_channel(isolate)
                isolates.append(isolate)
                couter += 1
        self._worker_num = workers
        return isolates

    async def run(self, multiprocessing: bool = False,
                  fast: bool = False,
                  ) -> None:
        """
        Run app.
        :param multiprocessing: if True, run every transport in separate processes; else run all in one process.
        :param fast:
        :return:
        """
        if not multiprocessing and fast:
            raise Exception("multiprocessing must be True if fast is True")

        cheat(self, AppNameAble, self.name)
        tasks: list[Coroutine[Any, Any, Any]] = []

        cheat(self, EnvAble, self._environment)

        cheat(self, ConfAble, self.conf)

        await self._environment.init()
        await call_signals(self, SignalType.BEFORE_APP_RUN, app=self)
        await self._environment.shutdown()

        if multiprocessing:
            self.manager.add_isolate_list(self._create_transport_isolates(workers=os.cpu_count() if fast else None))
            # tasks.extend(self.manager.perform())
            await self.manager.run()
        else:
            for transport in self._transports:
                united_socket = ISocket(self._transports[0].conf.socket.host, self._transports[0].conf.socket.port)
                tasks.extend(transport.perform(united_socket))

        if self.scheduler:
            self.channel.add_event_listener(TaskEvent.message_type, self.scheduler.add_task)
            # tasks.extend(self.scheduler.perform())
            await self.scheduler.run()

        # # TODO FIX:
        # if not multiprocessing:
        #     united_socket = ISocket(self._transports[0].conf.socket.host, self._transports[0].conf.socket.port)
        #     for transport in self._transports:
        #         tasks.extend(transport.perform(united_socket))

        tasks.extend(self.dispatcher.perform())
        tasks.append(self.channel.listen_consume())

        async def sig_handler():
            loguru.logger.info(
                f"{self.name} Received shutdown signal , exit")
            await call_signals(self, SignalType.AFTER_APP_STOP, app=self)
            self.scheduler.stop()
            self.manager.stop()
            exit(0)

        loop = asyncio.get_event_loop()
        loop.add_signal_handler(SIGINT, lambda: asyncio.create_task(sig_handler()))
        loop.add_signal_handler(SIGTERM, lambda: asyncio.create_task(sig_handler()))

        await self._environment.init()
        await asyncio.wait(tasks)
        await self._environment.shutdown()


#
# async def __return_metrics(self, event: MetricRequestEvent):
#     event = await self._metrics_store.on_metric_request(event)
#     await self._dispatcher.send_to_consume(event.sender, event)

def load_config(conf_path: Path, config_model: Type[GenConfig]) -> GenConfig:
    """
    Load environment config to user in
    :param conf_path:
    :param config_model: BaseModel to cast json file to pydantic
    :return: None
    """
    with open(conf_path, "r") as _json_file:
        conf = config_model(**json.loads(_json_file.read()))
        return conf



from __future__ import annotations

import asyncio
import io
import json
from dataclasses import dataclass, field
from typing import Dict, Any

import loguru
from sanic.response import BaseHTTPResponse

from web.kernel.types import RtWriter, RtMessage


@dataclass
class SseRtMessage(RtMessage):
    _SEPARATOR = "\r\n"
    event_name: str = field(default="")
    data: Dict | None = field(default=None)
    retry: int | None = field(default=None)

    def prepare(self, idx: int = 0, *args, **kwargs) -> str | bytes:
        buffer = io.StringIO()
        buffer.write('event: ' + self.event_name + self._SEPARATOR)
        if idx:
            buffer.write('id: ' + str(idx) + self._SEPARATOR)
        if self.retry:
            buffer.write('retry: ' + str(self.retry) + self._SEPARATOR)
        else:
            buffer.write("retry: " + "0" + self._SEPARATOR)
        if self.data:
            buffer.write('data: ' + json.dumps(self.data))
        else:
            buffer.write('data: {}')
        buffer.write("\r\n\r\n")
        return buffer.getvalue()

    @classmethod
    def ping_message(cls):
        return cls(event_name="ping").prepare()


class SseWriter(RtWriter):
    message_class = SseRtMessage
    obj: BaseHTTPResponse
    ping_enable: bool
    ping_timeout: int

    def __init__(self, obj: BaseHTTPResponse,
                 ping_enable: bool = False,
                 ping_timeout: int = 5):
        super().__init__(obj)
        self.ping_enable = ping_enable
        self.ping_timeout = ping_timeout

    async def write(self, message: Any) -> None:
        await self.obj.send(message)
        self.last_index += 1

    async def freeze(self):
        async with self:
            while True:
                if self.ping_enable:
                    await self.write(self.message_class.ping_message())
                await asyncio.sleep(self.ping_timeout)

    async def close(self, *args, **kwargs):
        await self.obj.send(end_stream=True)

    @staticmethod
    def content_type() -> str:
        return "text/event-stream; charset=utf-8"

    @staticmethod
    def default_headers() -> dict:
        return {"X-Accel-Buffering": "no"}

    async def __aenter__(self):
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        self.on_close()



from web.kernel.transport import Transport
from web.kernel.types import ConfAble, ISocket, TransportConf, AppNameAble

try:
    from grpclib.server import Server
    from grpclib.utils import graceful_exit
    from betterproto.grpc.grpclib_server import ServiceBase
except ImportError:
    raise ImportError("poetry install --with grpc")


class GrpcTransportConfig(TransportConf):
    pass


class GrpcTransport(Transport, AppNameAble, ConfAble[GrpcTransportConfig]):
    async def run(self, socket: ISocket):
        for service in self.services:
            if ServiceBase not in service.__class__.__mro__:
                raise RuntimeError(ServiceBase)
        server = Server(self.services)
        with graceful_exit([server]):
            await server.start(sock=socket)
            await server.wait_closed()



import asyncio
from typing import Coroutine, Callable

from loguru import logger

from web import settings
from web.kernel.types import GenIMessage, IChannel, IMessage, IMessageWaiter, EventListener, GenEventsType, \
    IMessageReturn


class Channel(IChannel):
    async def produce(self, msg: IMessage | IMessageReturn, need_answer: bool = False,
                      answer_timeout: float = 0) -> IMessage | None:
        if IMessageReturn not in msg.__class__.__mro__:
            self._inner_index += 1
            msg.inner_index = self._inner_index
        msg.sender = self.name
        if msg.exec_inner:
            if listeners := self._listeners.get(msg.message_type):
                for listener in listeners:
                    asyncio.create_task(listener(msg))
        self.produce_pipe.put(msg)
        if need_answer:
            waiter = IMessageWaiter(self, msg, timeout=answer_timeout)
            self.set_waiter(waiter)
            return await waiter.wait_on()

    async def sent_to_consume(self, msg: IMessage):
        self.consume_pipe.put(msg)

    async def listen_produce(self, callback: Callable[[GenIMessage], Coroutine]):
        while True:
            while self.produce_pipe.empty():
                await asyncio.sleep(0.01)
            r: GenIMessage = self.produce_pipe.get()
            if settings.DEBUG:
                logger.debug(
                    f"[{self.name}] Send: {r.message_type}")
            await callback(r)

    async def listen_consume(self):
        while True:
            while self.consume_pipe.empty():
                await asyncio.sleep(0.01)
            r: IMessage = self.consume_pipe.get()
            if settings.DEBUG:
                logger.debug(
                    f"[{self.name}] Receive: {r.message_type}")
            self.answer_to_waiters(r)
            callbacks = self._listeners.get(r.message_type)
            if not callbacks:
                continue
            for callback in callbacks:
                asyncio.create_task(callback(r))

    def add_event_listener(self, event_type: GenEventsType,
                           callback: EventListener, use_nested_classes: bool = False):
        def add():
            nonlocal self, event_type, callback, use_nested_classes
            if use_nested_classes and isinstance(event_type, str):
                raise AttributeError("Can't add_event_listener with use_nested_classes")

            def _add(event_name: str):
                nonlocal self
                nonlocal callback
                if event_name not in self._listeners:
                    self._listeners[event_name] = []
                self._listeners[event_name].append(callback)

            if isinstance(event_type, str):
                _add(event_type)
            else:
                def _raise(cls_type):
                    if not hasattr(cls_type, "message_type"):
                        raise AttributeError("Can't register listener cause message_type not found")

                if use_nested_classes:
                    for cls in event_type.__subclasses__():
                        _raise(cls)
                        _add(cls.message_type)
                else:
                    _raise(event_type)
                    _add(event_type.message_type)

        if isinstance(event_type, list):
            for i in event_type:
                add()
        else:
            add()



import asyncio
from dataclasses import dataclass, field

import loguru
from sanic.server.websockets.impl import WebsocketImplProtocol
from websockets.connection import OPEN

from web.kernel.types import RtMessage, RtWriter, RtReadCallback


@dataclass
class WsRtMessage(RtMessage):
    msg: str = field(default="")

    def prepare(self, *args, **kwargs) -> str | bytes:
        return self.msg

    @classmethod
    def ping_message(cls):
        return ""


class WsWriter(RtWriter[WsRtMessage]):
    read_callback: RtReadCallback
    obj: WebsocketImplProtocol

    def __init__(self, obj: WebsocketImplProtocol,
                 read_callback: RtReadCallback = None):
        super().__init__(obj)
        self.read_callback = read_callback

    async def write(self, message: str) -> None:
        await self.obj.send(message)

    async def freeze(self, ping_enable: bool = False, ping_timeout: int = 5):
        self.obj.connection_lost_waiter.add_done_callback(lambda s: self.on_close())
        if self.read_callback:
            async for msg in self.obj:
                await self.read_callback(self, msg)
        else:
            await self.obj.wait_for_connection_lost()

    async def close(self, code: int = 1000, reason: str = ""):
        await self.obj.close(code=code, reason=reason)



# from typing import TypeVar, Generic
#
#
# class T:
#     def w(self):
#         pass
#
#
# GenT = TypeVar("GenT", bound=T)
#
#
# class ChanAble(Generic[GenT]):
#     _chan: GenT | None
#
#     @property
#     def channel(self) -> GenT:
#         if self._chan:
#             return self._chan
#         else:
#             raise Exception("_channel not set")
#
#     @channel.setter
#     def channel(self, chan: GenT):
#         self._chan = chan
#
#
# class MyT(T):
#
#     def c(self):
#         pass
#
#
# GenM = TypeVar("GenT")
#
#
# class Test(Generic[GenM], ChanAble[GenM]):
#     pass
#
#
# if __name__ == '__main__':
#     Test[MyT]().channel.c()

# from _sanic import Sanic
#
# if __name__ == '__main__':
#     s = Sanic(name="Shit")
#     s.run(single_process=True)



import asyncio
from typing import Dict, Any, Coroutine

from web import settings
from web.kernel.messaging.channel import Channel
from web.kernel.messaging.channel import GenIMessage
from web.kernel.types import ChanAble, cheat, SettingsChangeEvent


async def on_settings_change(event: SettingsChangeEvent):
    setattr(settings, event.name, event.value)


class IDispatcher:
    channels: Dict[str, Channel]
    _msg_global_index: int
    state: Dict[str, Any]
    _master_channel: Channel

    def __init__(self):
        self.state = {}
        self.channels = {}
        self._events_listeners = {}
        self.collected_metrics = {}
        self._msg_global_index = 0

    def set_channel(self, chan_able: ChanAble, ext_name: str = None, master: bool = False) -> Channel:
        chan = Channel(len(self.channels) + 1, ext_name if ext_name else chan_able.name)
        chan.add_event_listener(SettingsChangeEvent, on_settings_change)
        chan_able.channel = chan
        cheat(chan_able, ChanAble, chan, nested=False)
        self.channels.update({chan.name: chan})
        if master:
            self._master_channel = chan
        return chan

    async def on_channel_sent(self, msg: GenIMessage):
        self._msg_global_index += 1
        msg.index = self._msg_global_index
        # asyncio.create_task(self.track_event(msg))
        if msg.destination != "__master__":
            asyncio.create_task(self._broadcast(msg))
        else:
            await self._master_channel.sent_to_consume(msg)

    async def _broadcast(self, msg: GenIMessage):
        for worker_name, ch in self.channels.items():
            if msg.sender == worker_name:
                continue
            if msg.destination in ["__all__", worker_name]:
                ch.consume_pipe.put(msg)

    async def send_to_consume(self, chan_name: str, event: GenIMessage):
        sender_channel = self.channels.get(chan_name)
        await sender_channel.produce(event)

    def perform(self, **kwargs) -> list[Coroutine[Any, Any, Any]]:
        tasks = []
        for channel in self.channels.values():
            tasks.append(channel.listen_produce(
                self.on_channel_sent
            ))
        return tasks

    # async def track_event(self, msg: GenericIMessage):
    #     listener = self.events_listeners.get(msg.message_type)
    #     if listener:
    #         await listener(msg)
    #
    #     if settings.METRICS_ENABLE:
    #         event_counter = CounterMetric("events_counter", self._msg_global_index)
    #         save_to = self.collected_metrics['__system_metric__']
    #         save_to.update({"events_counter": {"dispatcher_events_counter": event_counter}})
    #
    #         named_event_counter = save_to.get("named_events_counter")
    #         if not named_event_counter:
    #             mtr = CounterMetric("named_events_counter", value=1)
    #             mtr.add_label(event_name=msg.message_type)
    #             named_event_counter = {msg.message_type: mtr}
    #             save_to["named_events_counter"] = named_event_counter
    #         elif named_event_counter.get(msg.message_type):
    #             named_event_counter[msg.message_type].inc()
    #         else:
    #             mtr = CounterMetric("named_events_counter", value=1)
    #             mtr.add_label(event_name=msg.message_type)
    #             named_event_counter[msg.message_type] = mtr
    #         if settings.DEBUG:
    #             loguru.logger.debug(f'MetricsDispatcher - track event {msg}')



from __future__ import annotations

import os
import sys
from abc import ABC
from abc import ABCMeta
from dataclasses import dataclass
from functools import wraps
from traceback import extract_tb
from typing import Any, Callable, Coroutine, TypeVar, Generic, cast
from typing import Dict
from typing import List
from typing import Type

import loguru
from pydantic import ValidationError as PdValidationError, BaseModel as PdModel
from sanic import Request
from sanic import Sanic, file
from sanic import json, HTTPResponse
from sanic.errorpages import exception_response, JSONRenderer
from sanic.exceptions import BadRequest
from sanic.exceptions import SanicException
from sanic.handlers import ErrorHandler
from sanic.response import ResponseStream
from sanic.router import Router

from web import settings
from web.errors.app import ValidationError, ApplicationError
from web.kernel.transport import Transport
from web.kernel.types import DtoType, ConfAble
from web.kernel.types import Plugin, GenAsyncCall, ITransport
from web.trend.rest.utils.openapi_gen import add_openapi
from web.trend.rest.utils.openapi_spec import add_openapi_spec


class ServingPaths(PdModel):
    uri: str
    path: str
    route: bool = False


SanicStaticServing = dict[str, list[ServingPaths]] | None


class RouterConf(PdModel):
    openapi_filepath: str | None
    serving: SanicStaticServing


@dataclass
class PlugRouteConf:
    route: str
    method: str

    def __post_init__(self):
        self.method = self.method.lower()


class RestPlugin(Plugin, metaclass=ABCMeta):
    routes_conf: List[PlugRouteConf]
    required_attrs = ["routes_conf"]

    async def exec(self, handler: GenAsyncCall,
                   ctx: InputContext,
                   transport: ITransport,
                   prev_plugin_result: Any = None):
        raise NotImplementedError


def validate_dto(dto_cls: Type[PdModel] | None, request: Request) -> PdModel | None:
    if not dto_cls:
        return None
    if "json" in request.content_type:
        data = request.json
    elif "form" in request.content_type:
        data = {key: value[0] for key, value in request.form.items()}
    else:
        raise BadRequest("Incorrect content type")
    try:
        dto = dto_cls(**data)
        return dto
    except PdValidationError as ex:
        failed_fields = ex.errors()
        fields = [field["loc"][-1] for field in failed_fields]
        # commment_str = "Some of essential params failed : " + ", ".join(
        #     [field["loc"][-1] + " - " + field["msg"] for field in failed_fields])
        comment_str = f"""Some essential params failed: {", ".join([f'{field["loc"][-1]} - {field["msg"]}' for field in failed_fields])}"""
        context = {
            "fields": fields,
            "comment": comment_str
        }
        raise ValidationError(message=comment_str,
                              context=context)


@dataclass
class ProtectIdentity:
    pass


ProtectIdentityType = TypeVar("ProtectIdentityType", bound=ProtectIdentity)


@dataclass
class InputContext(Generic[DtoType, ProtectIdentityType]):
    request: Request
    dto: DtoType | None
    identity: ProtectIdentity | None
    r_kwargs: Dict


Protector = Callable[[Request, Transport], Coroutine[Any, Any, ProtectIdentityType | None]]
DtoValidator = Callable[[Type[DtoType], Request], PdModel]
HandlerType = Callable[[InputContext, Transport], Coroutine[Any, Any, HTTPResponse]]


class CustomRequestHandler(ABC):
    protector: Protector | None
    in_struct: Type[PdModel] | None
    validation_fnc: DtoValidator | None
    transport: Transport
    response_fabric: Callable[[Any], HTTPResponse] | None

    def __init__(self,
                 transport: Transport,
                 protector: Protector = None,
                 in_struct: Type[PdModel] | None = None,
                 validation_fnc: DtoValidator = validate_dto,
                 response_fabric: Callable[[Any], HTTPResponse] = json):
        self.protector = protector
        self.transport = transport
        self.in_struct = in_struct
        self.validation_fnc = validation_fnc
        self.response_fabric = response_fabric

    def __find_plugins(self, req) -> List[RestPlugin]:
        find_plugins = []
        for plug in self.transport.plugins:
            if RestPlugin in plug.__class__.__mro__:
                plug = cast(RestPlugin, plug)
                for route_cfg in plug.routes_conf:
                    if route_cfg == PlugRouteConf(req.route.path, req.method):
                        find_plugins.append(plug)
        return find_plugins

    def __call__(self, target: HandlerType):
        @wraps(target)
        async def f(*args, **kwargs):
            req: Request = args[0]
            if len(args) > 1:
                kwargs['ws'] = args[1]
            prot_identity = await self.protector(req, self.transport) if self.protector else None
            if self.in_struct:
                validated = self.validation_fnc(self.in_struct, req)
            else:
                validated = None
            incoming = InputContext(req, validated, prot_identity, kwargs)
            ret_val = None
            try:
                plugins = self.__find_plugins(req)
                if plugins:
                    if settings.DEBUG:
                        loguru.logger.debug(
                            f"[{self.transport.__class__.__name__}::RequestHandler]: find plugins: {plugins}")
                    for plug in plugins:
                        if settings.DEBUG:
                            loguru.logger.debug(
                                f"[{self.transport.__class__.__name__}::RequestHandler] Call plugin {plug.name}")
                        ret_val = await plug.exec(target, incoming, self.transport, prev_plugin_result=ret_val)
            except RuntimeError as e:
                if settings.DEBUG:
                    loguru.logger.warning(f"RequestHandler{e}")
            if not ret_val:
                ret_val = await target(incoming, self.transport)
            if isinstance(ret_val, (HTTPResponse, ResponseStream)):
                return ret_val
            return self.response_fabric(ret_val)

        return f


class CustomRouter(Router, ConfAble[RouterConf]):
    _router_conf: Dict
    open_api: bool

    def __init__(self,
                 routes_config: Dict,
                 open_api: bool = False):
        super().__init__()
        self._router_conf = routes_config
        self.open_api = open_api

    @classmethod
    def set_serving_config(cls, sanic: Sanic, config):
        """Configure serving static files"""
        if not hasattr(config, 'serving') or not config.serving:
            return
        for target, serv in config.serving.items():
            if not os.path.exists(target):
                os.makedirs(target, exist_ok=True)
            for serv_path in serv:
                if serv_path.path:
                    serv_path_target = os.path.join(target, serv_path.path)
                else:
                    serv_path_target = target
                if serv_path.route:
                    if not os.path.isfile(serv_path_target):
                        raise Exception("Serve static as route is possible only for one file")
                    else:
                        def get_static_route(path: str):
                            def static_route(r, *args, **kwargs):
                                return file(path)

                            return static_route

                        static_func = get_static_route(serv_path_target)
                        static_func.__name__ = "static_" + "_".join(serv_path_target.split('/'))
                        sanic.add_route(static_func, serv_path.uri)
                else:
                    if not os.path.exists(serv_path_target):
                        os.makedirs(serv_path_target, exist_ok=True)
                    sanic.static(serv_path.uri, file_or_directory=serv_path_target,
                                 name="static_" + "_".join(serv_path_target.split('/')))

    def set_routes(self, transport: Transport):
        version_prefix = self._router_conf.get("version_prefix")
        version_prefix = version_prefix if version_prefix else "/api/v"
        for endpoint, versions in self._router_conf.get("endpoints").items():
            if not isinstance(versions, dict):
                continue
            for version, params in versions.items():
                if not isinstance(params, dict):
                    continue
                methods_confs = []
                endpoint_handler = params.pop("handler", None)
                endpoint_protector = params.pop("protector", None)
                endpoint_response_fabric = params.pop("response_fabric", None)
                for method_name, method_params in params.items():
                    if not isinstance(method_params, dict):
                        continue
                    target_func = method_params.get('handler')
                    target_func = target_func if target_func else endpoint_handler
                    protector = method_params.get("protector")
                    protector = protector if protector else endpoint_protector
                    in_dto = method_params.get("in_dto")
                    out_dto = method_params.get("out_dto")
                    response_fabric = method_params.get("response_fabric")
                    response_fabric = response_fabric if response_fabric else endpoint_response_fabric

                    # handler = self.__find_handler_by_name(services, method_conf.handler, prefix=method_prefix)
                    if response_fabric:
                        chain = CustomRequestHandler(transport,
                                                     protector=protector,
                                                     in_struct=in_dto,
                                                     response_fabric=response_fabric)(target_func)
                    else:
                        chain = CustomRequestHandler(transport,
                                                     protector=protector,
                                                     in_struct=in_dto)(target_func)
                    if self.open_api:
                        chain = add_openapi_spec(uri=endpoint,
                                                 method_name=method_name,
                                                 func=target_func,
                                                 handler=chain,
                                                 in_dto=in_dto,
                                                 out_dto=out_dto)
                    if method_name.lower() == 'websocket':
                        self.ctx.app.add_websocket_route(
                            handler=chain,
                            uri=endpoint,
                            name=f"__{method_name}_{chain.__name__}",
                            version=version,
                            version_prefix=version_prefix
                        )
                    else:
                        chain.__name__ = f"__{method_name}_{chain.__name__}"
                        self.ctx.app.add_route(
                            uri=endpoint,
                            methods={method_name.upper()},
                            handler=chain,
                            name=f"__{method_name}_{chain.__name__}",
                            version=version,
                            version_prefix=version_prefix
                        )

        try:
            if self.conf:
                self.set_serving_config(self.ctx.app, self.conf)

                if hasattr(self.conf, "openapi_filepath") and self.conf.openapi_filepath:
                    self.ctx.app.config.SWAGGER_UI_CONFIGURATION = {
                        "docExpansion": 'none'
                    }
                    add_openapi(self.ctx.app, self.conf.openapi_filepath)
                else:
                    self.ctx.app.config.OAS = False
        except RuntimeError:
            pass


class StatusCodes:
    APPLICATION_CODE = 409
    REPORT_CODE = 501


class CustomJSONRenderer(JSONRenderer):
    def _generate_output(self, *, full):
        output = {
            "description": self.title,
            "status": self.status,
            # "message": self.text,
            "error": getattr(self.exception, "context", {})
        }
        if not output['error'].get('type'):
            output['error']['type'] = getattr(self.exception, "error_type", "C")

        if not output['error'].get('class'):
            output['error']['class'] = getattr(self.exception, "default_class", 999)

        if not output['error'].get('subclass'):
            output['error']['subclass'] = getattr(self.exception, "default_subclass", 999)

        if not output['error'].get('code'):
            output['error'][
                'code'] = f"{output['error']['type']}-{output['error']['class']}-{output['error']['subclass']}"

        if self.text:
            if output.get('error'):
                if not output['error'].get('comment'):
                    output['error'].update({"comment": self.text})
            else:
                output['error'] = {"comment": self.text}

        if full:
            _, exc_value, __ = sys.exc_info()
            exceptions = []

            while exc_value:
                exceptions.append(
                    {
                        "type": exc_value.__class__.__name__,
                        "exception": str(exc_value),
                        "frames": [
                            {
                                "file": frame.filename,
                                "line": frame.lineno,
                                "name": frame.name,
                                "src": frame.line,
                            }
                            for frame in extract_tb(exc_value.__traceback__)
                        ],
                    }
                )
                exc_value = exc_value.__cause__

            output["path"] = self.request.path
            output["args"] = self.request.args
            output["exceptions"] = exceptions[::-1]

        return output


class CustomErrorHandler(ErrorHandler):
    def default(self, request, exception):
        """Convert ApplicationError to SanicException"""
        renderer = None
        if isinstance(exception, ApplicationError):
            if not isinstance(exception, SanicException):
                exception = type(exception.__class__.__name__,
                                 (SanicException,),
                                 {"message": exception.message if exception.message else "",
                                  "status_code": StatusCodes.APPLICATION_CODE,
                                  "error_type": exception.error_type,
                                  "default_class": exception.default_class,
                                  "default_subclass": exception.default_subclass, })(context=exception.context)
            renderer = CustomJSONRenderer
        else:
            self.log(request, exception)
        fallback = request.app.config.FALLBACK_ERROR_FORMAT
        return exception_response(
            request,
            exception,
            debug=settings.DEBUG,
            base=self.base,
            fallback=fallback,
            renderer=renderer
        )



from __future__ import annotations

import os
from abc import ABCMeta
from pathlib import Path
from typing import List, Any, Dict, Tuple

import loguru

from web import settings
from web.kernel.messaging.channel import GenIMessage
from web.kernel.proc.isolate import Isolate
from web.kernel.types import ISocket, RtWriter, \
    GenEventsType, RtResolver, BackTaskTarget, BackTaskTrigger, BackTaskCallback, TaskReturnEvent, \
    TaskEvent, ITransport, PluginContainer, GenAsyncCall, call_signals, SignalType


class Transport(ITransport, metaclass=ABCMeta):
    async def new_rt(self, writer: RtWriter,
                     events_type: GenEventsType,
                     resolver: RtResolver,
                     use_nested_events: bool = False):
        if writer.__class__ not in self.rt_writers.keys():
            self.rt_writers[writer.__class__] = []
        self.rt_writers[writer.__class__].append(writer)

        async def send_in_writer(message: GenIMessage):
            if msg := await resolver(writer, message):
                await writer.write(msg.prepare(writer.last_index))

        self.channel.add_event_listener(events_type, send_in_writer,
                                        use_nested_classes=use_nested_events)

        async def on_disconnect():
            self.rt_writers[writer.__class__].remove(writer)

        await writer.with_lost_callback(on_disconnect).run()

    async def back_task(self,
                        target: BackTaskTarget,
                        args=None,
                        kwargs=None,
                        trigger: BackTaskTrigger = None,
                        exec_type: str = "default",
                        on_error: BackTaskCallback = None,
                        on_complete: BackTaskCallback = None,
                        need_result: bool = False
                        ) -> TaskReturnEvent | None:
        if kwargs is None:
            kwargs = {}
        if args is None:
            args = []
        if not settings.SCHEDULER_ENABLE:
            raise ImportError("Apscheduler is not installed on your system. scheduler not enable")
        else:
            return await self.channel.produce(TaskEvent(
                target,
                trigger=trigger,
                args=args,
                kwargs=kwargs,
                exec_type=exec_type,
                on_error=on_error,
                on_complete=on_complete
            ), need_answer=need_result)

    async def set_plugin(self, file_path: Path) -> Tuple[str, str | Exception]:
        container = PluginContainer(file_path)
        status, message = container.import_plugin()
        if status == "success":
            await self._plugins_store.set(container.plugin.name, container)
        if settings.DEBUG:
            getattr(loguru.logger, status)(f"[{self.__class__.__name__}::{self._plugins_store}::{container}] {message}")
        return status, message

    async def remove_plugin(self, plugin_name: str):
        await self._plugins_store.remove(plugin_name)


class TransportIsolate(Isolate):
    _transport: Transport
    _socket: ISocket

    def __init__(self, name: str, transport: Transport, socket: ISocket, **kwargs):
        super().__init__(name, **kwargs)
        self._transport = transport
        self._socket = socket

    async def work(self, *args, **kwargs) -> None:
        self._transport.init_plugins()
        await call_signals(self._transport, SignalType.BEFORE_TRANSPORT_WORK, self._transport)
        loguru.logger.info(f"[{self.name}::Isolate(pid={os.getpid()})]: socket {self._socket}")
        await self._transport.run(self._socket)



import logging

from loguru import logger
from pydantic import BaseModel


class FileLoggerSettings(BaseModel):
    log_file: str
    rotation_trigger: str  # "500 MB"  "12:00" "1 week" "10 days"
    compression: str  # "zip"
    retention: str


class LogsHandler(logging.Handler):
    def __init__(self, file_log_settings: FileLoggerSettings | None):
        super().__init__()
        if file_log_settings:
            logger.add(file_log_settings.log_file, rotation=file_log_settings.rotation_trigger,
                       compression=file_log_settings.compression, retention=file_log_settings.retention)

    def emit(self, record):
        logger_opt = logger.opt(depth=10, exception=record.exc_info)
        try:
            logger_opt.log(record.levelname, record.getMessage())
        except Exception as exx:
            logger.warning(f"LOGGER {record.name} ERROR {exx}")
            pass


def setup_loggers(file_log_settings: FileLoggerSettings | None = None):
    handler = LogsHandler(file_log_settings)
    for logger_name, logger_obj in logging.root.manager.loggerDict.items():
        logging.getLogger(logger_name).handlers.clear()
        logging.getLogger(logger_name).handlers = [handler]
        # logging.basicConfig(handlers=[handler], level=0, force=True)



import multiprocessing
from multiprocessing import Manager
from typing import List, Coroutine, Any, Dict

import loguru

from web import settings
from web.kernel.proc.isolate import Isolate
from web.kernel.types import EnvAble, cheat, ConfAble


class ProcManager(EnvAble, ConfAble):
    _worker_num: int
    _isolates: Dict[str, Isolate]
    _manager: Manager

    def __init__(self):
        self._manager = Manager()
        self._worker_num = 0
        self._isolates = {}

    def perform(self) -> list[Coroutine[Any, Any, Any] | Coroutine[Any, Any, None]]:
        return [self.run()]

    def _add_isolate(self, isolate: Isolate, run: bool = False):
        self._isolates.update({isolate.name: isolate})
        cheat(isolate, EnvAble, self.env)
        if run:
            ctx = multiprocessing.get_context("fork")
            isolate.fork(ctx)
        if settings.DEBUG:
            loguru.logger.debug(f"[{self}] Add isolate {isolate.name}")

    def add_isolate_list(self, isolates: List[Isolate], run: bool = False):
        for i in isolates:
            self._add_isolate(i, run)

    def add_isolate(self, isolate: Isolate, run: bool = False):
        self._add_isolate(isolate, run)

    def remove_isolate(self, name: str):
        self._isolates.pop(name)

    async def run(self):
        with self._manager as manager:
            ctx = multiprocessing.get_context("fork")
            for isolate in self._isolates.values():
                if not isolate.process:
                    isolate.fork(ctx)

    def stop(self):
        for i in self._isolates.values():
            i.process.terminate()

    @property
    def isolates(self):
        return self._isolates

    def __repr__(self):
        return f"{self.__class__.__name__}"
